@@ -114,8 +114,8 @@ impl NbtStructureError {
     pub(crate) fn type_mismatch(expected: &'static str, found: &'static str) -> Self {
         NbtStructureError {
             repr: NbtStructureErrorRepr::TypeMismatch {
-                expected: Box::new(expected),
-                found: Box::new(found),
+                expected,
+                found,
             },
         }
     }
@@ -154,10 +154,9 @@ enum NbtStructureErrorRepr {
         index: usize,
         length: usize,
     },
-    // Keep the size of this type down to that of a wide pointer
     TypeMismatch {
-        expected: Box<&'static str>,
-        found: Box<&'static str>,
+        expected: &'static str,
+        found: &'static str,
     },
 }
 
@@ -167,6 +166,10 @@ enum NbtStructureErrorRepr {
 /// as one as well.
 ///
 /// [`NbtCompound`]: crate::tag::NbtCompound
+#[deprecated(
+    since = "0.2.3",
+    note = "This trait will eventually be made obsolete with serde compatibility"
+)]
 pub trait NbtRepr: Sized {
     /// Updates the data in this type based on the given compound. The intention is that data is copied, not
     /// moved, from the compound to update this type.
