@@ -2,15 +2,9 @@
 
 mod assets;
 use assets::*;
-use quartz_nbt::{
-    compound,
-    io::{self, Flavor},
-    serde::{deserialize, serialize, Array},
-    NbtCompound,
-    NbtList,
-};
-use serde::{Deserialize, Serialize};
-use std::{collections::HashMap, io::Cursor};
+use quartz_nbt::{NbtCompound, NbtList, NbtTag, compound, io::{self, Flavor}, serde::{Array, deserialize, deserialize_from, deserialize_from_buffer, serialize}};
+use serde::{de::Visitor, ser::SerializeStruct, Deserialize, Serialize};
+use std::{collections::HashMap, convert::TryInto, io::{Cursor, Seek, SeekFrom}};
 
 #[derive(Serialize, Deserialize, PartialEq)]
 struct Level {
@@ -330,6 +324,27 @@ fn enum_serde() {
         .unwrap()
         .0;
     assert_eq!(deserialized_struct, test_struct);
+
+    #[derive(PartialEq, Deserialize, Debug)]
+    struct DeserializeByName {
+        enums: Vec<A>
+    }
+
+    let deserialize_by_name = DeserializeByName {
+        enums: vec![A::G, A::E, A::F]
+    };
+
+    let deserialize_by_name_nbt = compound! {
+        "enums": ["G", "E", "F"]
+    };
+
+    let mut serialized_struct = Cursor::new(Vec::<u8>::new());
+    io::write_nbt(&mut serialized_struct, None, &deserialize_by_name_nbt, Flavor::Uncompressed).unwrap();
+    serialized_struct.seek(SeekFrom::Start(0)).unwrap();
+
+    let deserialized_struct: DeserializeByName = deserialize_from(&mut serialized_struct, Flavor::Uncompressed).unwrap().0;
+
+    assert_eq!(deserialize_by_name, deserialized_struct);
 }
 
 #[test]
@@ -340,6 +355,10 @@ fn array_serde() {
         byte_array2: Array<Vec<i8>>,
         int_array: Array<Vec<i32>>,
         long_array: Array<Vec<i64>>,
+        prim_byte_array: Array<[u8; 3]>,
+        prim_byte_array2: Array<[i8; 3]>,
+        prim_int_array: Array<[i32; 3]>,
+        prim_long_array: Array<[i64; 3]>,
     }
 
     let test_struct = Foo {
@@ -347,6 +366,10 @@ fn array_serde() {
         byte_array2: vec![51, 32, 99].into(),
         int_array: vec![120, 99999, 12].into(),
         long_array: vec![2122, 121212, 6666666].into(),
+        prim_byte_array: [0u8, 1, 2].into(),
+        prim_byte_array2: [-1i8, -10, -128].into(),
+        prim_int_array: [5i32, -10, 20].into(),
+        prim_long_array: [i64::MIN, i64::MAX, 0].into(),
     };
 
     let serialized_struct = serialize(&test_struct, None, Flavor::Uncompressed).unwrap();
@@ -362,7 +385,11 @@ fn array_serde() {
         "byte_array": [B; 12, 13, 14],
         "byte_array2": [B; 51, 32, 99],
         "int_array": [I; 120, 99999, 12],
-        "long_array": [L; 2122, 121212, 6666666]
+        "long_array": [L; 2122, 121212, 6666666],
+        "prim_byte_array": [B; 0, 1, 2],
+        "prim_byte_array2": [B; -1, -10, -128],
+        "prim_int_array": [I; 5, -10, 20],
+        "prim_long_array": [L; i64::MIN, i64::MAX, 0]
     };
 
     assert_eq!(struct_nbt, validation_nbt);
@@ -371,6 +398,167 @@ fn array_serde() {
         .unwrap()
         .0;
     assert_eq!(deserialized_struct, test_struct);
+
+    // Wrapper to ensure we test the `serialize_bytes` method
+    struct Bytes<'a>(&'a [u8]);
+
+    impl<'a> Serialize for Bytes<'a> {
+        fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+        where S: serde::Serializer {
+            serializer.serialize_bytes(&self.0)
+        }
+    }
+
+    // Wrapper to ensure that we test the `deserialize_byte_buf` method
+    struct ByteBuf(Vec<u8>);
+
+    impl<'de> Deserialize<'de> for ByteBuf {
+        fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
+        where D: serde::Deserializer<'de> {
+            struct ByteBufVisitor;
+
+            impl<'de> Visitor<'de> for ByteBufVisitor {
+                type Value = Vec<u8>;
+
+                fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
+                    write!(formatter, "A byte buf")
+                }
+
+                fn visit_byte_buf<E>(self, v: Vec<u8>) -> Result<Self::Value, E>
+                where E: serde::de::Error {
+                    Ok(v)
+                }
+            }
+
+            Ok(ByteBuf(deserializer.deserialize_byte_buf(ByteBufVisitor)?))
+        }
+    }
+
+    // Large arrays aren't incorporated into the serde data model by default, so we need to handle
+    // them manually
+    #[derive(PartialEq, Eq, Debug)]
+    struct SerAsSlice {
+        large_byte_array: Box<[u8; 1024]>,
+        large_int_array: Box<[i32; 256]>,
+    }
+
+    impl Serialize for SerAsSlice {
+        fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+        where S: serde::Serializer {
+            let mut serialize_struct = serializer.serialize_struct("SerAsSlice", 2)?;
+            let bytes: &[u8] = &*self.large_byte_array;
+            // Ensure that we call the `serialize_bytes` method
+            serialize_struct.serialize_field("large_byte_array", &Bytes(bytes))?;
+            let slice: &[i32] = &*self.large_int_array;
+            let arr: Array<&[i32]> = Array::from(slice);
+            serialize_struct.serialize_field("large_int_array", &arr)?;
+            serialize_struct.end()
+        }
+    }
+
+    // Quick and dirty implementation since we test structs elsewhere
+    impl<'de> Deserialize<'de> for SerAsSlice {
+        fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
+        where D: serde::Deserializer<'de> {
+            struct SerAsSliceVisitor;
+
+            impl<'de> Visitor<'de> for SerAsSliceVisitor {
+                type Value = SerAsSlice;
+
+                fn expecting(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+                    write!(f, "Things")
+                }
+
+                fn visit_map<A>(self, mut map: A) -> Result<Self::Value, A::Error>
+                where A: serde::de::MapAccess<'de> {
+                    let mut large_byte_array: Option<Box<[u8; 1024]>> = None;
+                    let mut large_int_array: Option<Box<[i32; 256]>> = None;
+
+                    // Quick and dirty implementation since we test structs elsewhere
+                    while let Some(key) = map.next_key::<String>()? {
+                        match key.as_str() {
+                            "large_byte_array" => {
+                                assert!(
+                                    large_byte_array.is_none(),
+                                    "large_byte_array deserialized twice"
+                                );
+                                let bytes: ByteBuf = map.next_value()?;
+                                large_byte_array = Some(Box::new(
+                                    bytes
+                                        .0
+                                        .try_into()
+                                        .expect("large_byte_array has incorrect length"),
+                                ));
+                            }
+                            "large_int_array" => {
+                                assert!(
+                                    large_int_array.is_none(),
+                                    "large_int_array deserialized twice"
+                                );
+                                let arr: Vec<i32> = map.next_value()?;
+                                large_int_array = Some(Box::new(
+                                    arr.try_into()
+                                        .expect("large_int_array has incorrect length"),
+                                ));
+                            }
+                            _ => panic!("Invalid key: {}", key),
+                        }
+                    }
+
+                    Ok(SerAsSlice {
+                        large_byte_array: large_byte_array.expect("large_byte_array missing"),
+                        large_int_array: large_int_array.expect("large_int_array missing"),
+                    })
+                }
+            }
+
+            let ser_as_slice = deserializer.deserialize_struct(
+                "SerAsSlice",
+                &["large_byte_array", "large_int_array"],
+                SerAsSliceVisitor,
+            )?;
+            Ok(ser_as_slice)
+        }
+    }
+
+    let large_byte_array: Vec<_> = (0 .. 1024).map(|i| ((i * 7 + 13) % 255) as u8).collect();
+    let large_int_array: Vec<_> = (0 .. 256).collect();
+
+    let ser_as_slice = SerAsSlice {
+        large_byte_array: Box::new(large_byte_array.clone().try_into().unwrap()),
+        large_int_array: Box::new(large_int_array.clone().try_into().unwrap()),
+    };
+
+    let serialized_struct = serialize(&ser_as_slice, None, Flavor::Uncompressed).unwrap();
+
+    let struct_nbt = io::read_nbt(
+        &mut Cursor::new(serialized_struct.clone()),
+        Flavor::Uncompressed,
+    )
+    .unwrap()
+    .0;
+
+    let validation_nbt = compound! {
+        "large_byte_array": large_byte_array,
+        "large_int_array": large_int_array
+    };
+
+    // Ensure the validation nbt is valid
+    assert!(matches!(
+        validation_nbt.get::<_, &NbtTag>("large_byte_array"),
+        Ok(NbtTag::ByteArray(_))
+    ));
+    assert!(matches!(
+        validation_nbt.get::<_, &NbtTag>("large_int_array"),
+        Ok(NbtTag::IntArray(_))
+    ));
+
+    assert_eq!(struct_nbt, validation_nbt);
+
+    let deserialized_struct: SerAsSlice = deserialize(&serialized_struct, Flavor::Uncompressed)
+        .unwrap()
+        .0;
+    assert_eq!(deserialized_struct, ser_as_slice);
 }
 
 #[test]
@@ -383,12 +571,7 @@ fn vec_serde() {
         tuple: Vec<Tuple>,
         enumeration: Vec<Enumeration>,
         mixed_enumeration: Vec<Enumeration>,
-        bar_arr: Array<Vec<Bar>>,
-        strings_arr: Array<Vec<String>>,
-        baz_arr: Array<Vec<Baz>>,
-        tuple_arr: Array<Vec<Tuple>>,
-        nested_arr: Vec<Array<Vec<Array<Vec<i8>>>>>,
-        enumeration_arr: Array<Vec<Enumeration>>,
+        nested_arr: Vec<Vec<Array<Vec<i8>>>>,
         enum_of_vec: Enumeration,
         empty_byte_array: Array<Vec<i8>>,
         empty_int_array: Array<Vec<i32>>,
@@ -422,21 +605,16 @@ fn vec_serde() {
     }
 
     let test_struct = Foo {
-        baz_arr: vec![Baz(52)].into(),
         baz: vec![Baz(99), Baz(42), Baz(88)],
         bar: vec![Bar { a: 32 }, Bar { a: 99 }],
-        tuple_arr: vec![Tuple(12, 12, 12)].into(),
         tuple: vec![Tuple(343, 89, 102), Tuple(33, 897, 457)],
         strings: vec!["test".to_owned(), "test test test".to_owned()],
         enumeration: vec![Enumeration::A, Enumeration::B, Enumeration::E],
         mixed_enumeration: vec![Enumeration::D(12), Enumeration::F { a: 14 }],
-        bar_arr: vec![Bar { a: 35 }].into(),
-        strings_arr: vec!["tteesstt".to_owned()].into(),
         nested_arr: vec![
             vec![vec![1, 20, 9].into()].into(),
             vec![vec![3, 5, 10].into(), vec![99, 10, 32].into()].into(),
         ],
-        enumeration_arr: vec![Enumeration::A, Enumeration::C, Enumeration::E].into(),
         enum_of_vec: Enumeration::G(vec![vec![Bar { a: 13 }, Bar { a: 9 }], vec![Bar { a: 14 }]]),
         empty_byte_array: Vec::new().into(),
         empty_int_array: Vec::new().into(),
@@ -471,12 +649,6 @@ fn vec_serde() {
             [343i16, 89i16, 102i16],
             [33i16, 897i16, 457i16]
         ],
-        "bar_arr": [
-            {
-                "a": 35i32
-            }
-        ],
-        "strings_arr": ["tteesstt"],
         "enumeration": [0i32, 1i32, 4i32],
         "mixed_enumeration": [
             {
@@ -488,10 +660,6 @@ fn vec_serde() {
                 }
             }
         ],
-        "baz_arr": [B; 52],
-        "tuple_arr": [
-            [12i16, 12i16, 12i16]
-        ],
         "nested_arr": [
             [
                 [B; 1, 20, 9]
@@ -501,7 +669,6 @@ fn vec_serde() {
                 [B; 99, 10, 32]
             ]
         ],
-        "enumeration_arr": [I; 0, 2, 4],
         "enum_of_vec": {
             "G": [
                 [
@@ -519,14 +686,13 @@ fn vec_serde() {
                 ]
             ]
         },
-        // Make sure all empty lists coerce to the same type
-        "empty_byte_array": [],
-        "empty_int_array": [],
-        "empty_long_array": [],
+        "empty_byte_array": [B;],
+        "empty_int_array": [I;],
+        "empty_long_array": [L;],
         "empty_tag_list": [],
-        "seq_empty_byte_array": [[], []],
-        "seq_empty_int_array": [[], []],
-        "seq_empty_long_array": [[], []],
+        "seq_empty_byte_array": [[B;], [B;]],
+        "seq_empty_int_array": [[I;], [I;]],
+        "seq_empty_long_array": [[L;], [L;]],
         "seq_empty_tag_list": [[], []],
     };
 
@@ -603,3 +769,121 @@ fn option_serde() {
         .0;
     assert_eq!(deserialized_struct, test_struct);
 }
+
+#[test]
+fn borrowed_serde() {
+    #[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
+    struct BorrowedData<'a> {
+        bytes: Array<&'a [u8]>,
+        string: &'a str,
+    }
+
+    const BYTES: &[u8] = &[1, 2, 3, 4, 5];
+
+    let data = BorrowedData {
+        bytes: Array::from(BYTES),
+        string: "is my unsafe code sound?",
+    };
+
+    let serialized_struct = serialize(&data, None, Flavor::Uncompressed).unwrap();
+    let struct_nbt = io::read_nbt(
+        &mut Cursor::new(serialized_struct.clone()),
+        Flavor::Uncompressed,
+    )
+    .unwrap()
+    .0;
+
+    let validation_nbt = compound! {
+        "bytes": BYTES.to_vec(),
+        "string": "is my unsafe code sound?"
+    };
+
+    assert_eq!(struct_nbt, validation_nbt);
+
+    let deserialized_struct: BorrowedData<'_> =
+        deserialize_from_buffer(&serialized_struct).unwrap().0;
+    assert_eq!(deserialized_struct, data);
+
+    // Just to make it clear what the drop order is
+    drop(deserialized_struct);
+    drop(serialized_struct);
+}
+
+#[test]
+fn inlined_nbt() {
+    #[derive(Debug, PartialEq, Serialize, Deserialize)]
+    struct Inlined {
+        byte: NbtTag,
+        short: NbtTag,
+        int: NbtTag,
+        long: NbtTag,
+        float: NbtTag,
+        double: NbtTag,
+        byte_array: NbtTag,
+        string: NbtTag,
+        tag_list: NbtTag,
+        tag_compound: NbtTag,
+        int_array: NbtTag,
+        long_array: NbtTag,
+        list: NbtList,
+        compound: NbtCompound,
+        normal_field: i32,
+    }
+
+    let test_compound = SNBT_EDGE_CASES_VALIDATE.clone();
+    let test_list = test_compound
+        .get::<_, &NbtList>("chaotic_array")
+        .unwrap()
+        .clone();
+
+    let inlined = Inlined {
+        byte: NbtTag::Byte(31),
+        short: NbtTag::Short(-214),
+        int: NbtTag::Int(328795),
+        long: NbtTag::Long(-928592323532),
+        float: NbtTag::Float(2.71828),
+        double: NbtTag::Double(-3.14159),
+        byte_array: NbtTag::ByteArray(vec![-1, 2, -3, 4]),
+        string: NbtTag::String("foobar".to_owned()),
+        tag_list: NbtTag::List(test_list.clone()),
+        tag_compound: NbtTag::Compound(test_compound.clone()),
+        int_array: NbtTag::IntArray(vec![-1_000_000, 2_000_000]),
+        long_array: NbtTag::LongArray(Vec::new()),
+        list: test_list.clone(),
+        compound: test_compound.clone(),
+        normal_field: 42,
+    };
+
+    let serialized_struct = serialize(&inlined, None, Flavor::Uncompressed).unwrap();
+    let struct_nbt = io::read_nbt(
+        &mut Cursor::new(serialized_struct.clone()),
+        Flavor::Uncompressed,
+    )
+    .unwrap()
+    .0;
+
+    let validation_nbt = compound! {
+        "byte": 31i8,
+        "short": -214i16,
+        "int": 328795i32,
+        "long": -928592323532i64,
+        "float": 2.71828f32,
+        "double": -3.14159f64,
+        "byte_array": [B; -1, 2, -3, 4],
+        "string": "foobar",
+        "tag_list": test_list.clone(),
+        "tag_compound": test_compound.clone(),
+        "int_array": [I; -1_000_000, 2_000_000],
+        "long_array": [L;],
+        "list": test_list,
+        "compound": test_compound,
+        "normal_field": 42i32
+    };
+
+    assert_eq!(struct_nbt, validation_nbt);
+
+    let deserialized_struct: Inlined = deserialize(&serialized_struct, Flavor::Uncompressed)
+        .unwrap()
+        .0;
+    assert_eq!(deserialized_struct, inlined);
+}
